;##################################################################
; Favorites Manager
;##################################################################

#include "FavManager.h"

/*
teststruct.txt is a test file, just put it in the %AppData%\FavMan folder and rename it Favorites.txt
or if you like just create your own by right clicking in the box and adding folders and urls
*/

/*
DEBUGLIB equ \Programming\RadASM\GoAsm\lib\Debug.lib
DEBUGEXE equ "\Programming\RadAsm\GoAsm\lib\dbgwin.exe"

#include "\Programming\RadASM\GoAsm\inc\debug.a"
DBGWIN_DEBUG_ON = 1 // use this to enable Spy/StopSpy/DbgDump and TrapException
;DBGWIN_SHOWONLYERRORS = 0 // setting this to 1 will suppress PrintError if no error exists 
*/

MAX_FILESIZE equ 131072 // 128KB

FAVENTRY STRUCT
	iType	DB		// 1 = folder, 0 = entry
	pad		DB	3 DUP (?) // Pad to 4 bytes (maybe for future use)
	Label	DB 	0
ENDS

.DATA
	hInstance			HANDLE 	?
	hDlg				HANDLE	?

	hDragIml			HANDLE	?
	TVDragItem			HANDLE	?
	IsDragging			BOOL	?

	FavFolderPath		CHAR	"%AppData%\FavMan",0
	ALIGN 4
	DefFileName			CHAR	"Favorites.txt",0
	ALIGN 4
	FavFilePath			CHAR	MAX_PATH	DUP (?)
	AppDataPath			CHAR	MAX_PATH	DUP (?)

.CODE

Start:
	invoke GetModuleHandle, 0
	mov [hInstance],EAX
	invoke InitCommonControls

	invoke ExpandEnvironmentStrings,offset FavFolderPath,offset AppDataPath,MAX_PATH
	invoke PathIsDirectory,offset AppDataPath
	test eax,eax
	jnz >
		invoke SHCreateDirectoryEx,NULL,offset AppDataPath,NULL
	:

	invoke lstrcpy,offset FavFilePath,offset AppDataPath
	invoke PathAppend,offset FavFilePath,offset DefFileName

	invoke DialogBoxParam,[hInstance],1000,0,ADDR DlgProc,0
	invoke ExitProcess,0

DlgProc FRAME hwnd,uMsg,wParam,lParam
	uses ebx
	LOCAL pt	:POINT

	.WM_INITDIALOG
		cmp D[uMsg],WM_INITDIALOG
		jne >>.WM_COMMAND
			invoke GetDlgItem,[hwnd],1001
			mov ebx,eax
			invoke GetWindowLong,ebx,GWL_WNDPROC
			invoke SetWindowLong,ebx,GWL_USERDATA,eax
			invoke SetWindowLong,ebx,GWL_WNDPROC,offset FavouritesSubClass
			mov eax,[hwnd]
			mov [hDlg],eax
			xor eax,eax
			inc eax
			ret

	.WM_COMMAND
		cmp D[uMsg],WM_COMMAND
		jne >.WM_NOTIFY
		cmp W[wParam],1002
		jne >>.EXIT
			invoke GetDlgItem,[hwnd],1001
			invoke LoadFavorites,eax,offset FavFilePath
			xor eax,eax
			inc eax
			ret

	.WM_NOTIFY
		cmp D[uMsg],WM_NOTIFY
		jne >>.WM_LBUTTONUP
			mov eax,[lParam]
			cmp D[eax+NMHDR.code],TVN_BEGINDRAG
			jne >>.EXIT
				mov D[IsDragging],TRUE
				invoke TVBeginDrag,[eax+NMHDR.hwndFrom],[lParam]
				xor eax,eax
				inc eax
				ret

	.WM_LBUTTONUP
		cmp D[uMsg],WM_LBUTTONUP
		jne >.WM_MOUSEMOVE
			cmp D[IsDragging],TRUE
			jne >>.EXIT
				mov D[IsDragging],FALSE
				invoke GetDlgItem,[hwnd],1001
				invoke TVEndDrag,eax
				xor eax,eax
				inc eax
				ret

	.WM_MOUSEMOVE
		cmp D[uMsg],WM_MOUSEMOVE
		jne >.WM_CONTEXTMENU
			cmp D[IsDragging],TRUE
			jne >.EXIT
				invoke GetCursorPos,ADDR pt
				invoke ImageList_DragMove,[pt.x],[pt.y]
				xor eax,eax
				inc eax
				ret

	.WM_CONTEXTMENU
		cmp D[uMsg], WM_CONTEXTMENU
		jne >.WM_CLOSE
			invoke GetDlgItem,[hwnd],1001
			invoke ShowContextMenu,eax
			xor eax,eax
			inc eax
			ret

	.WM_CLOSE
		cmp D[uMsg],WM_CLOSE
		jne >.DEFPROC
		invoke EndDialog,[hwnd],0

	.DEFPROC
		mov eax,FALSE
		RET

	.EXIT

	mov eax, TRUE
	RET
ENDF

/*
This part of the code is the subclassing routine for
the treeview, it handles messages that must be processed
by the control itself.
*/

FavouritesSubClass FRAME hwnd,uMsg,wParam,lParam
	
	// This subclassing is used to clear the treeview of any buffers
	// when it is destroyed

	.WM_DESTROY
	cmp D[uMsg],WM_DESTROY
	jne >>.EXIT
		// Don't attempt to save or clear an empty treeview
		invoke SendMessage,[hwnd],TVM_GETCOUNT,0,0
		test eax,eax
		jz >>.EXIT
		invoke SaveFavorites,[hwnd],offset AppDataPath,offset DefFileName, 0
		// free memory allocated for urls before we destroy the window
		invoke SendMessage,[hwnd],TVM_GETNEXTITEM,TVGN_ROOT,TVI_ROOT
		invoke ClearFavorites,[hwnd], eax

	.EXIT
	
	invoke GetWindowLong,[hwnd],GWL_USERDATA
	invoke CallWindowProc,eax,[hwnd],[uMsg],[wParam],[lParam]
	RET
ENDF

/*
This part of the code loads and saves the favorites file
as well as populating the treeview control
*/

ClearFavorites FRAME hTreeView,hItem
	LOCAL tvi	:TVITEM
	
	// NOTE this is a recursive algorithm that walks through the treeview

	// Does the item have children ?
	mov eax,[hItem]
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_CHILDREN+TVIF_PARAM
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi

	// If there are no children then it is a url, free the memory buffer
	cmp D[tvi.cChildren], FALSE
	je >
		// If there are children then recurse through them
		// It is possible for an item to be flagged as having children
		// when there are actually none
		invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CHILD,[hItem]
		test eax,eax
		jz >
		invoke ClearFavorites,[hTreeView],eax
	:
	cmp D[tvi.lParam],0
	je >
	invoke GlobalFree,[tvi.lParam]
	:
	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_NEXT,[hItem]
	test eax,eax
	jz >
		invoke ClearFavorites,[hTreeView],eax
	:
	RET
ENDF

LoadFavorites FRAME hTreeView,pFileName
	LOCAL hFile				:%HANDLE
	LOCAL cbBytesRead		:%DWORD32
	LOCAL FileSize			:%DWORD64
	LOCAL pBuffer			:%PTR
	LOCAL cchText			:%DWORD32
	LOCAL hParent			:%HANDLE

	invoke SendMessage,[hTreeView],TVM_EXPAND,TVE_COLLAPSERESET + TVE_COLLAPSE,TVI_ROOT
	invoke SendMessage,[hTreeView],TVM_GETIMAGELIST,TVSIL_NORMAL,0
	test eax,eax
	jnz >
		invoke ImageList_LoadImage,[hInstance],500,16,10,0FFFFFFh,IMAGE_BITMAP,LR_CREATEDIBSECTION
		invoke SendMessage,[hTreeView],TVM_SETIMAGELIST,TVSIL_NORMAL,eax
	:

	// Read the entire file into memory
	invoke CreateFile,[pFileName],GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL
	mov [hFile],eax
	test eax,eax
	js >>.ERRORCREATE

	invoke GetFileSizeEx,[hFile],offset FileSize

	// Calculate the next 1KB boundary above FileSize
	mov eax,[FileSize]
	add eax,1024
	and eax,-1024

	// Create a buffer large enough to hold the entire file
	invoke VirtualAlloc,NULL,eax,MEM_COMMIT,PAGE_READWRITE
	mov [pBuffer],eax

	// Read the file into the buffer
	invoke ReadFile,[hFile],[pBuffer],[FileSize],ADDR cbBytesRead,NULL

	// We're done with the file so close it
	invoke CloseHandle,[hFile]

	// Verify that the entire file was read
	mov eax,[FileSize]
	cmp [cbBytesRead],eax
	jne >>.ERRORREAD

	invoke ParseFile,[hTreeView],[pBuffer]

	.ERRORREAD
	invoke VirtualFree,[pBuffer],NULL,MEM_RELEASE
	.ERRORCREATE

	RET
ENDF

ParseFile FRAME hTreeView , pFileData
	uses edi,esi,ebx
	LOCAL mbi		:MEMORY_BASIC_INFORMATION
	LOCAL hParent	:%HANDLE
	LOCAL pParents	:%PTR
	LOCAL junk		:%DWORD32

	invoke GlobalAlloc,GMEM_FIXED,1024

	mov [pParents],eax
	mov D[eax],TVI_ROOT
	:
	// Calculate the end of the allocated buffer
	invoke VirtualQuery,[pFileData],offset mbi,SIZEOF MEMORY_BASIC_INFORMATION

	mov eax,[pFileData]
	and eax,0FFFh ; assumes a page size of 4096 (0x1000)
	sub [mbi.RegionSize],eax
	mov ecx,[mbi.RegionSize]

	// Read the current line
	mov edi,[pFileData]
	mov esi,edi
	mov al,10
	repne scasb

	// Get rid of the CR/LF
	mov B[edi-1],0
	mov B[edi-2],0

	// edi holds a pointer to the start of the next line
	// esi holds a pointer to the start of the current line
	
	// Check to see if the next item is a child
	// This is done by checking the number of tabs at the
	// start of the next line and comparing it to the current
	// line

	// The number of tabs tells us the parent, the first parent is always 0

	mov ebx,0
	.1
	mov al,[esi+ebx]
	inc ebx
	cmp al,9
	je <.1
	dec ebx
	// Multiply the number of tabs by 4 to get the 
	// offset in the array of parent handles
	shl ebx,2

	mov edx,[pParents]

	mov D[hParent],TVI_ROOT
	test ebx,ebx
	jz >.2
		mov eax,ebx
		sub eax,4
		mov eax,[edx+eax]
		mov [hParent],eax
	.2

	// OK we have everything we need to add the item
	// we just have to parse the line
	
	// the first digit represents the type of entry (folder or url)

	mov edx,ebx
	shr edx,2

	mov eax,esi
	add eax,FAVENTRY.Label
	add eax,edx
	mov ecx,[esi+edx]
	// we're only concerned with the last digit in the ascii code (0=30h,1=31h)
	and ecx,0Fh

	mov edx,NULL
	cmp ecx,0
	jne >.3
		// This is a url so we need to extract the actual URL
		// If there is no label it will replace the label with the actual url
		push eax
		// Parse the label and URL
		mov bl,[eax]
			.4
				inc eax
				mov bl,[eax]
				cmp bl,","
				jne <.4
				mov B[eax],0
				mov edx,eax
				inc edx
				pop eax
	.3

	invoke TviAddItem,[hTreeView],[hParent],ecx,eax,edx

	// Insert this item as the last parent, just in case it
	// ends up having children
	mov edx,[pParents]
	mov [edx+ebx],eax

	mov [pFileData],edi
	xor eax,eax
	mov al,[edi]
	test eax,eax
	jnz <<

	invoke GlobalFree,[pParents]

	RET
ENDF

TviAddItem Frame hTreeView, hParent, itype, pszName, pszData
	LOCAL	tvins			:TVINSERTSTRUCT

	mov eax,[hParent]
	mov [tvins.hParent],eax

	mov D[tvins.hInsertAfter],0
	mov D[tvins.itemex.mask],TVIF_TEXT + TVIF_PARAM + TVIF_IMAGE + TVIF_SELECTEDIMAGE + TVIF_CHILDREN
	mov eax,[pszName]
	mov [tvins.itemex.pszText],eax

	mov D[tvins.itemex.lParam],NULL
	cmp D[pszData],NULL
	je >
		invoke lstrlen,[pszData]
		inc eax
		push eax
		// Create a buffer to hold the url data
		invoke GetProcessHeap
		pop ecx
		invoke HeapAlloc,eax,HEAP_ZERO_MEMORY,ecx
		invoke lstrcpy,eax,[pszData]
		mov [tvins.itemex.lParam],eax
	:

	mov eax,[itype]
	mov D[tvins.itemex.cChildren],eax
	mov D[tvins.itemex.iImage],eax
	mov D[tvins.itemex.iSelectedImage],eax
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvins

	RET
ENDF

SaveFavorites FRAME hTreeView, pFilePath, pFilename, bSilent
	LOCAL hFile	:%HANDLE
	LOCAL ofn :OPENFILENAME
	LOCAL openName[MAX_PATH]:%CHAR
	
	invoke lstrcpy,offset openName,[pFilename]

	cmp D[bSilent],TRUE
	je >>.WRITEFILE

	invoke RtlZeroMemory,offset ofn, SIZEOF OPENFILENAME

	mov D[ofn.lStructSize], SIZEOF OPENFILENAME
	mov eax, [hDlg]
	mov [ofn.hwndOwner], eax
	mov eax, [hInstance]
	mov [ofn.hInstance], eax
	lea eax, openName
	mov [ofn.lpstrFile], eax
	mov D[ofn.nMaxFile], MAX_PATH

	mov eax,[pFilePath]
	mov [ofn.lpstrInitialDir],eax
	mov D[ofn.lpstrTitle], offset szSaveCaption
	mov D[ofn.Flags], OFN_HIDEREADONLY

	invoke GetSaveFileName,offset ofn
	test eax,eax
	jz >
	.WRITEFILE
	invoke CreateFile,offset openName,GENERIC_WRITE,FILE_SHARE_WRITE,NULL,CREATE_ALWAYS,NULL,NULL
	mov [hFile],eax
	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_ROOT,TVI_ROOT
	invoke WalkFavorites,[hTreeView],[hFile],eax,0
	invoke CloseHandle,[hFile]
	:
	RET
	szSaveCaption: CHAR	"Save favorites file",0
ENDF

WalkFavorites FRAME hTreeView, hFile, hItem, iLevel
	LOCAL tvi	:TVITEM
	LOCAL buffer[64]:%CHAR

	// NOTE this is a recursive algorithm that walks through the treeview
	// Since for this walk we need to extract the text we also need a buffer
	// to store it during the recursion. Since the buffer must be unique for
	// each level of the recursion it has to be local. This may cause problems
	// with extremely deep nesting but I can't figure out another way around it.
	// For now labels will be limited to 64 bytes

	// Does the item have children ?
	lea eax,buffer
	mov [tvi.pszText],eax
	mov D[tvi.cchTextMax],64
	mov eax,[hItem]
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_CHILDREN+TVIF_PARAM+TVIF_TEXT
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi

	// iLevel holds the number of tabs
	// buffer holds the label
	// tvi.Children holds the type
	// tvi.lParam holds the url buffer address

	// build our line and save it to the file
	invoke BuildFileLine,[hFile],[iLevel],[tvi.cChildren],offset buffer,[tvi.lParam]

	// If there are no children then skip to the next one
	// otherwise recurse the search down a level
	cmp D[tvi.cChildren], FALSE
	je >
		// We are about to recurse to a child item so increment the level
		inc D[iLevel]
		// If there are children then step through them
		invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CHILD,[hItem]
		test eax,eax
		jz >.DOWN
		invoke WalkFavorites,[hTreeView],[hFile],eax,[iLevel]
		.DOWN
		// Done with the recursion so decrement the level
		dec D[iLevel]
	:

	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_NEXT,[hItem]
	test eax,eax
	jz >
		invoke WalkFavorites,[hTreeView],[hFile],eax,[iLevel]
	:
	RET
ENDF

BuildFileLine FRAME hFile,iLevel, iType, pszText, pszUrl
	uses ebx,esi,edi
	LOCAL FileLine[4096]:%CHAR
	LOCAL szTabs[1024]:%CHAR
	LOCAL szType[8]:%CHAR
	LOCAL cbWritten:%DWORD32
	
	invoke RtlZeroMemory,offset FileLine,2048
	invoke RtlZeroMemory,offset szTabs,2048
	mov D[szType],20202020h
	mov D[szType+4],0

	mov eax,[iType]
	and eax,0Fh
	add eax,30h
	mov B[szType],al

	// First set the number of tabs
	lea edi,szTabs
	mov ebx,[iLevel]
	test ebx,ebx
	jz >.EXIT
	:
		dec ebx
		mov B[edi+ebx],9
		test ebx,ebx
		jnz <
	.EXIT

	cmp D[iType],0
	jne >
	invoke wsprintf,offset FileLine,offset Format1,offset szTabs,offset szType,[pszText],[pszUrl]
	add esp,24
	invoke WriteFile,[hFile],offset FileLine,eax,offset cbWritten,NULL
	RET
	:
	invoke wsprintf,offset FileLine,offset Format2,offset szTabs,offset szType,[pszText]
	add esp,20
	invoke WriteFile,[hFile],offset FileLine,eax,offset cbWritten,NULL
	RET
	Format1:	CHAR	"%s%s%s,%s",13,10,0
	Format2:	CHAR	"%s%s%s",13,10,0

ENDF

/*
This part of the code handles the item dragging
*/

TVBeginDrag FRAME hTreeView,lParam
	LOCAL DragStart		:POINT
	LOCAL tvi			:TVITEM
	LOCAL rect			:RECT

	mov eax,[lParam]
	mov eax,[eax+NM_TREEVIEW.itemNew.hItem]
	mov [TVDragItem],eax
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_STATE
	mov D[tvi.state],TVIS_DROPHILITED
	invoke SendMessage,[hTreeView],TVM_SETITEM,0,ADDR tvi

	invoke GetCursorPos,ADDR DragStart
	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_DROPHILITE,[TVDragItem]

	invoke SendMessage,[hTreeView],TVM_CREATEDRAGIMAGE,0,[TVDragItem]
	mov [hDragIml],eax

	invoke ImageList_BeginDrag,[hDragIml],0,-8,-8
	invoke GetDesktopWindow
	invoke ImageList_DragEnter,eax,[DragStart.x],[DragStart.y]

	invoke GetParent,[hTreeView]
	invoke SetCapture,eax

	ret
ENDF

TVEndDrag FRAME hTreeView
	uses ebx
	LOCAL pt					:POINT
	LOCAL tvi					:TVITEM
	LOCAL tvht					:TVHITTESTINFO
	LOCAL buffer[MAX_PATH]		:%CHAR
	LOCAL hNewParent			:%HANDLE

	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_DROPHILITE,NULL

	// release the mouse capture
	invoke ReleaseCapture
	
	// end the drag operation and destroy the drag image list
	invoke GetDesktopWindow
	invoke ImageList_DragLeave,eax
	invoke ImageList_EndDrag
	invoke ImageList_Destroy,[hDragIml]

	// Find the closest item to the cursor
	invoke GetCursorPos,ADDR tvht.pt
	invoke ScreenToClient,[hTreeView],ADDR tvht.pt
	invoke SendMessage,[hTreeView],TVM_HITTEST,0,ADDR tvht

	// Get some info about the item at the cursor
	// Get the item's parent
	// If the item is a folder then the new parent will be the folder itself
	
	mov eax,[tvht.hItem]
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_CHILDREN
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi
	
	cmp D[tvi.cChildren],FALSE
	je >
		mov eax,[tvht.hItem]
		mov [hNewParent],eax
		jmp >>.PARENTFOUND

	:
		invoke SendMessage, [hTreeView],TVM_GETNEXTITEM,TVGN_PARENT,[tvht.hItem]
		mov [hNewParent], eax

	.PARENTFOUND

	// Is this a root item ?
	test eax,eax
	jnz >.FINDPOSITION
		// This is a root item so pass TVI_ROOT as the new parent
		mov D[hNewParent], TVI_ROOT
	.FINDPOSITION

	// If the drop target is outside the client area then do not move the item
	mov ebx, [tvht.flags]
	cmp ebx, TVHT_ABOVE
	jne >>.TVHT_BELOW
		ret
	
	.TVHT_BELOW
	cmp ebx, TVHT_BELOW
	jne >>.TVHT_TOLEFT
		ret
	
	.TVHT_TOLEFT
	cmp ebx, TVHT_TOLEFT
	jne >>.TVHT_TORIGHT
		ret
	
	.TVHT_TORIGHT
	cmp ebx, TVHT_TORIGHT
	jne >>.MOVEITEM
		ret
	
	.MOVEITEM // insert the item above the item's parent

	invoke MoveFavorite,[hTreeView],[TVDragItem],[hNewParent],[tvht.hItem]
	ret

	.EXIT
	ret
ENDF

MoveFavorite FRAME hTreeView,hItem,hParent,hInsertAfter
	LOCAL tvis			:TVINSERTSTRUCT
	LOCAL szBuffer[64]	:%CHAR
	LOCAL hNewItem		:%HANDLE

	mov eax,[hItem]
	mov [tvis.itemex.hItem],eax
	mov D[tvis.itemex.mask],TVIF_TEXT+TVIF_IMAGE+TVIF_PARAM+TVIF_CHILDREN+TVIF_SELECTEDIMAGE
	mov eax,OFFSET szBuffer
	mov D[tvis.itemex.pszText],eax
	mov D[tvis.itemex.cchTextMax],64
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvis.itemex
	cmp D[tvis.itemex.iImage],FALSE
	jne >>.MOVEFOLDER // HAVE TO MOVE CHILDREN WITH FOLDERS !!!!

	mov eax,[hParent]
	mov [tvis.hParent],eax
	mov eax,[hInsertAfter]
	mov D[tvis.hInsertAfter],eax
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvis

	// the item label is not always visible if we don't allow the
	// treeview time to redraw itself before selecting the item
	push eax
	invoke Sleep,100
	pop eax
	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_CARET,eax

	invoke SendMessage,[hTreeView],TVM_DELETEITEM,0,[hItem]
	RET
	
	.MOVEFOLDER
	// First create the new folder
	mov eax,[hParent]
	mov [tvis.hParent],eax
	mov eax,[hInsertAfter]
	mov D[tvis.hInsertAfter],eax
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvis
	mov [hNewItem],eax

	// Get the first child
	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CHILD,[hItem]
	// If there is no child skip the move
	test eax,eax
	jz >
	// Move the children
	invoke MoveChildren,[hTreeView],eax,[hNewItem]
	:

	// delete the original folder and all its children
	invoke SendMessage,[hTreeView],TVM_DELETEITEM,0,[hItem]

	invoke Sleep,100
	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_CARET,[hNewItem]
	.EXIT
	RET
ENDF

MoveChildren FRAME hTreeView, hItem, hParent
	LOCAL tvis	:TVINSERTSTRUCT
	LOCAL buffer[64]:%CHAR
	LOCAL hNewItem:%HANDLE

	// recurse through the children and move them one at a time

	// Does the item have children ?
	lea eax,buffer
	mov [tvis.item.pszText],eax
	mov D[tvis.item.cchTextMax],64
	mov eax,[hItem]
	mov [tvis.item.hItem],eax
	mov D[tvis.item.mask],TVIF_TEXT+TVIF_IMAGE+TVIF_PARAM+TVIF_CHILDREN+TVIF_SELECTEDIMAGE
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvis.item

	// Create the item
	mov eax,[hParent]
	mov [tvis.hParent],eax
	mov D[tvis.hInsertAfter],TVI_LAST
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvis
	mov [hNewItem],eax

	// If there are no children then skip to the next one
	// otherwise recurse the search down a level
	cmp D[tvis.item.cChildren], FALSE
	je >
		// If there are children then step through them
		invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CHILD,[hItem]
		test eax,eax
		jz >
		invoke MoveChildren,[hTreeView],eax,[hNewItem]
	:

	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_NEXT,[hItem]
	test eax,eax
	jz >
		invoke MoveChildren,[hTreeView],eax,[hParent]
	:
	RET
ENDF

/*
This part of the code handles adding and removing favorites and folders
*/

AddUrlDialog FRAME hwnd,uMsg,wParam,lParam
	LOCAL urltext[2048]:%CHAR
	LOCAL labeltext[64]:%CHAR

	.WM_INITDIALOG
		cmp D[uMsg],WM_INITDIALOG
		jne >.WM_COMMAND
			// lParam is used to pass the handle to the treeview
			// we'll just store it in a control to avoid using a
			// global variable.
			invoke GetDlgItem,[hwnd],1001
			invoke SetWindowLong,eax,GWL_USERDATA,[lParam]
			jmp >>.EXIT

	.WM_COMMAND
		cmp D[uMsg],WM_COMMAND
		jne >>.WM_CLOSE
		cmp W[wParam],1001
		jne >>
			invoke SendDlgItemMessage,[hwnd],1003,WM_GETTEXT,2048,offset urltext
			invoke SendDlgItemMessage,[hwnd],1006,WM_GETTEXT,64,offset labeltext
			invoke GetDlgItem,[hwnd],1001
			invoke GetWindowLong,eax,GWL_USERDATA
			invoke AddFavorite,eax,offset labeltext,offset urltext
			invoke PostMessage,[hwnd],WM_CLOSE,0,0
			xor eax,eax
			ret
		:
		cmp D[wParam],1002
		jne >>.EXIT
			invoke PostMessage,[hwnd],WM_CLOSE,0,0
			xor eax,eax
			ret

	.WM_CLOSE
		cmp D[uMsg],WM_CLOSE
		jne >>.EXIT
			invoke EndDialog,[hwnd],0

	.EXIT
		xor eax,eax
		RET
ENDF

AddFolderDialog FRAME hwnd,uMsg,wParam,lParam
	LOCAL urltext[2048]:%CHAR
	LOCAL labeltext[64]:%CHAR

	.WM_INITDIALOG
		cmp D[uMsg],WM_INITDIALOG
		jne >.WM_COMMAND
			// lParam is used to pass the handle to the treeview
			// we'll just store it in a control to avoid using a
			// global variable.
			invoke GetDlgItem,[hwnd],1001
			invoke SetWindowLong,eax,GWL_USERDATA,[lParam]
			invoke GetDlgItem,[hwnd],1003
			invoke EnableWindow,eax,FALSE
			jmp >>.EXIT

	.WM_COMMAND
		cmp D[uMsg],WM_COMMAND
		jne >>.WM_CLOSE
		cmp W[wParam],1001
		jne >>
			invoke SendDlgItemMessage,[hwnd],1006,WM_GETTEXT,64,offset labeltext
			invoke GetDlgItem,[hwnd],1001
			invoke GetWindowLong,eax,GWL_USERDATA
			invoke AddFolder,eax,offset labeltext
			invoke PostMessage,[hwnd],WM_CLOSE,0,0
			xor eax,eax
			ret
		:
		cmp D[wParam],1002
		jne >>.EXIT
			invoke PostMessage,[hwnd],WM_CLOSE,0,0
			xor eax,eax
			ret

	.WM_CLOSE
		cmp D[uMsg],WM_CLOSE
		jne >>.EXIT
			invoke EndDialog,[hwnd],0

	.EXIT
		xor eax,eax
		RET
ENDF

AddFavorite FRAME hTreeView, pszLabel, pszData
	LOCAL tvi		:TVITEM
	LOCAL tvht		:TVHITTESTINFO
	LOCAL tvis		:TVINSERTSTRUCT
	LOCAL hNewParent:%HANDLE

	// get the current item from the treeview
	// Find the closest item to the cursor
	invoke GetCursorPos,ADDR tvht.pt
	invoke ScreenToClient,[hTreeView],ADDR tvht.pt
	invoke SendMessage,[hTreeView],TVM_HITTEST,0,ADDR tvht
	test eax,eax
	jz >>.ITEMISROOT

	// Is it a folder ?
	mov eax,[tvht.hItem]
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_IMAGE
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi

	cmp D[tvi.iImage],1
	jne >>.ITEMISURL
		// Add the url to the currently selected folder
		mov eax,[tvht.hItem]
		mov [hNewParent], eax
		jmp >>.ADDITEM

	.ITEMISURL
		invoke SendMessage, [hTreeView],TVM_GETNEXTITEM,TVGN_PARENT,[tvht.hItem]
		mov [hNewParent], eax
		jmp >>.ADDITEM

	.ITEMISROOT
		mov D[hNewParent], TVI_ROOT

	.ADDITEM

	invoke lstrlen,[pszData]
	inc eax
	push eax
	// Create a buffer to hold the url data
	invoke GetProcessHeap
	pop ecx
	invoke HeapAlloc,eax,HEAP_ZERO_MEMORY,ecx
	invoke lstrcpy,eax,[pszData]

	mov [tvis.item.lParam],eax
	mov D[tvis.item.mask],TVIF_TEXT + TVIF_PARAM + TVIF_IMAGE
	mov eax,[pszLabel]
	mov [tvis.item.pszText],eax
	mov D[tvis.item.iImage],0
	mov eax,[hNewParent]
	mov [tvis.hParent],eax
	mov eax,[tvht.hItem]
	mov D[tvis.hInsertAfter],eax
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvis
	// the item label is not always visible if we don't allow the
	// treeview time to redraw itself before selecting the item
	push eax
	invoke Sleep,100
	pop eax
	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_CARET,eax
	RET
ENDF

AddFolder FRAME hTreeView, pszLabel
	LOCAL tvi		:TVITEM
	LOCAL tvht		:TVHITTESTINFO
	LOCAL tvis		:TVINSERTSTRUCT
	LOCAL hNewParent:%HANDLE

	// get the current item from the treeview
	// Find the closest item to the cursor
	invoke GetCursorPos,ADDR tvht.pt
	invoke ScreenToClient,[hTreeView],ADDR tvht.pt
	invoke SendMessage,[hTreeView],TVM_HITTEST,0,ADDR tvht
	test eax,eax
	jz >>.ITEMISROOT

	// Is it a folder ?
	mov eax,[tvht.hItem]
	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_IMAGE
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi

	cmp D[tvi.iImage],1
	jne >>.ITEMISURL
		// Add the url to the currently selected folder
		mov eax,[tvht.hItem]
		mov [hNewParent], eax
		jmp >>.ADDITEM

	.ITEMISURL
		invoke SendMessage, [hTreeView],TVM_GETNEXTITEM,TVGN_PARENT,[tvht.hItem]
		mov [hNewParent], eax
		jmp >>.ADDITEM

	.ITEMISROOT
		mov D[hNewParent], TVI_ROOT

	.ADDITEM

	mov D[tvis.item.lParam],NULL
	mov D[tvis.item.mask],TVIF_TEXT + TVIF_PARAM + TVIF_IMAGE + TVIF_CHILDREN + TVIF_SELECTEDIMAGE
	mov eax,[pszLabel]
	mov [tvis.item.pszText],eax
	mov D[tvis.item.cChildren],1
	mov D[tvis.item.iImage],1
	mov D[tvis.item.iSelectedImage],1
	mov eax,[hNewParent]
	mov [tvis.hParent],eax
	mov eax,[tvht.hItem]
	mov D[tvis.hInsertAfter],eax
	invoke SendMessage,[hTreeView],TVM_INSERTITEM,0,OFFSET tvis

	push eax
	invoke Sleep,100
	pop eax
	invoke SendMessage,[hTreeView],TVM_SELECTITEM,TVGN_CARET,eax
	RET
ENDF

RemoveItem FRAME hTreeView
	LOCAL hItem		:%HANDLE
	LOCAL tvi		:TVITEM
	LOCAL buffer[64]:%CHAR
	LOCAL msg[128]	:%CHAR

	// get the currently selected item
	invoke SendMessage, [hTreeView],TVM_GETNEXTITEM,TVGN_CARET,NULL
	mov [hItem],eax

	mov [tvi.hItem],eax
	mov D[tvi.mask],TVIF_TEXT + TVIF_CHILDREN + TVIF_PARAM
	lea eax,buffer
	mov [tvi.pszText],eax
	mov D[tvi.cchTextMax],64
	invoke SendMessage,[hTreeView],TVM_GETITEM,0,offset tvi

	invoke wsprintf,offset msg,"Delete %s, are you sure ?",offset buffer
	add esp,12
	invoke MessageBox,[hTreeView],offset msg,"Delete favorite",MB_ICONQUESTION + MB_YESNO
	cmp eax,IDYES
	jne >.EXIT
	cmp D[tvi.cChildren],TRUE
	jne >
	invoke MessageBox,[hTreeView],"This item may have contents, continue ?","Delete favorite",MB_ICONWARNING + MB_OKCANCEL
	cmp eax,IDCANCEL
	je >.EXIT
	invoke SendMessage, [hTreeView],TVM_GETNEXTITEM,TVGN_CHILD,[hItem]
	invoke ClearFavorites,[hTreeView],eax
	:
	cmp D[tvi.lParam],0
	je >
		invoke GetProcessHeap
		invoke HeapFree,eax,NULL,[tvi.lParam]
	:
	invoke SendMessage,[hTreeView],TVM_DELETEITEM,0,[hItem]
	.EXIT
	RET
ENDF

EditItemLabel FRAME hTreeView
	LOCAL hEdit		:%HANDLE
	LOCAL hItem		:%HANDLE
	LOCAL buffer[64]:%CHAR

	invoke SetFocus,[hTreeView]
	invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CARET,NULL
	mov [hItem],eax
	test eax,eax
	jz >>.EXIT
		// The thing about TVM_EDITLABEL is that it requires that the
		// treeviews parent handle the notifications. By subclassing
		// the edit control we can avoid this. Ofcourse there are always
		// trade offs, if you do it this way you cannot abort an edit
		// operation. Or at least I couldn't figure out how...
		invoke SendMessage,[hTreeView],TVM_EDITLABEL,0,[hItem]
		test eax,eax
		jz >>.EXIT
		mov [hEdit],eax
		invoke SetWindowLong,[hEdit],GWL_WNDPROC,offset EditSubClass
		invoke SetWindowLong,[hEdit],GWL_USERDATA,eax
		invoke SendMessage,[hEdit],EM_LIMITTEXT,63,0
	.EXIT
	RET
ENDF

EditSubClass FRAME hwnd,uMsg,wParam,lParam
	LOCAL buffer[64]:%CHAR
	LOCAL hTreeView	:%HANDLE
	LOCAL tvi		:TVITEM

	.WM_DESTROY
	cmp D[uMsg],WM_DESTROY
	jne >>.EXIT
		invoke GetParent,[hwnd]
		mov [hTreeView],eax
		invoke GetWindowText,[hwnd],offset buffer,63

		invoke SendMessage,[hTreeView],TVM_GETNEXTITEM,TVGN_CARET,NULL
		mov [tvi.hItem],eax
		mov D[tvi.mask],TVIF_TEXT
		lea eax,buffer
		mov [tvi.pszText],eax
		mov D[tvi.cchTextMax],64
		invoke SendMessage,[hTreeView],TVM_SETITEM,0,offset tvi

	.EXIT
	invoke GetWindowLong,[hwnd],GWL_USERDATA
	invoke CallWindowProc,eax,[hwnd],[uMsg],[wParam],[lParam]

	RET
ENDF

/*
This part of the code displays and tracks the context sensitive menu
*/

ShowContextMenu FRAME hTreeView
	uses ebx
	LOCAL hPopMenu  :%HANDLE
	LOCAL pt		:POINT
	LOCAL MCommand	:%HANDLE
	LOCAL hWin		:%HANDLE
	LOCAL tvi		:TVITEM
	LOCAL tvht		:TVHITTESTINFO

	invoke GetParent,[hTreeView]
	mov [hWin],eax

	invoke GetCursorPos, ADDR pt
	invoke ScreenToClient,[hWin],ADDR pt
	invoke ChildWindowFromPoint, [hWin], [pt.x], [pt.y]
	cmp eax,[hTreeView]
	jnz >>.EXIT
		// Get some info about the item
		invoke GetCursorPos,ADDR tvht.pt
		invoke ScreenToClient,[hTreeView],ADDR tvht.pt
		invoke SendMessage,[hTreeView],TVM_HITTEST,0,ADDR tvht
		// Save the result in EBX to use later
		mov ebx,eax

		invoke GetCursorPos, ADDR pt
		invoke CreatePopupMenu
		mov [hPopMenu], eax
		invoke AppendMenu, [hPopMenu], MF_STRING, 100, "Add favorite"
		invoke AppendMenu, [hPopMenu], MF_STRING, 101, "Add folder"
		// EBX holds the result of the hittest call, if its 0 then the cursor was not on an item
		// In that case we do not show the delete menu selection
		test ebx,ebx
		jz >
		invoke AppendMenu, [hPopMenu], MF_STRING, 102, "Delete"
		invoke AppendMenu, [hPopMenu], MF_STRING, 103, "Edit label"
		:
		invoke TrackPopupMenu, [hPopMenu], TPM_LEFTALIGN + TPM_RETURNCMD,\
				[pt.x], [pt.y], NULL, [hTreeView], NULL
		mov [MCommand],eax

		mov eax,[tvht.hItem]
		test eax,eax
		js >>.DONE
		cmp D[MCommand],100 // Add url
		jnz >L1
			invoke DialogBoxParam,[hInstance],2000,[hWin],offset AddUrlDialog,[hTreeView]
			jmp >>.DONE

		L1:
		cmp D[MCommand],101 // Add folder
		jnz >L2
			invoke DialogBoxParam,[hInstance],2000,[hWin],offset AddFolderDialog,[hTreeView]
			jmp >>.DONE

		L2:
		cmp D[MCommand],102 // delete
		jnz >L3
			invoke RemoveItem,[hTreeView]
			jmp >>.DONE

		L3:
		cmp D[MCommand],103 // edit label
		jnz >>.DONE
			invoke EditItemLabel,[hTreeView]
			jmp >>.DONE

	.DONE
	invoke DestroyMenu, [hPopMenu]
	
	.EXIT
	ret

ENDF
